      program pctfr
      
      implicit none
      
      character(len=256) :: infile, new_name

#ifdef MPI
      include 'mpif.h'
#endif

      include '../include/sizes.inc'
      include '../include/allprogs.inc'
      include '../include/info.inc'
      include '../include/com_01.inc'
      include '../include/mem_time.inc'
      include '../include/pifhead.inc'

      logical :: confine_3D = .false.
      real :: zero=0.0
      
C     Local variables, used only in the main routine
      real center(2)

      integer inunit, mcw
      integer first_iq, last_iq, max_iq_slab,
     +     first_z,last_z, first_hk,last_hk, max_hk_slab, num_hkl,
     +     den_min_x, den_max_x, den_min_y, den_max_y, 
     +     den_min_z, den_max_z, mem_size, 
     +     min_xyz, max_xyz, hkl_max, max_z_slab,
     +     first_rr_out, last_rr_out, max_rr_out_slab,
     +     first_u, last_u, max_u_slab,
     +     first_y, last_y, max_y_slab

      integer nrffts_1_row, nrffts_1_col, nrffts_1_omega,
     +     nrffts_2_row, nrffts_2_col, nrffts_2_omega
      
      integer alloc_stat, ierr, err_code

      double precision elapsed_time
      integer clock1, clock2, clock3, clock4, ios
            
      real, allocatable :: ctf_val(:,:), abs_ctf(:,:),
     $     disef(:,:), map3d(:,:,:), rffta_omega(:)
      real, allocatable :: pixels(:)

      integer, allocatable   :: indx(:)
      real, allocatable      :: theta(:), phi(:), omega(:)
      real, allocatable      :: x(:), y(:), mag(:), score(:)
      real, allocatable      :: delf_maj_ptle(:), delf_min_ptle(:)
      real, allocatable      :: ang_maj_ptle(:)
      character, allocatable :: orient_flag(:), origin_flag(:)
      complex, allocatable   :: fft3d(:,:,:), fft_klh(:)
      character(len=256), allocatable :: linein(:)

      character(len=18) :: subname = 'PCTFR main routine'
      
      real mapprjbig, enlargefact
      real fmin, fmax, favg, fstd
      integer fft_ann, nptles
      integer i, j, k, count, id, img_id
      
      progname = 'PCTFR'

#ifdef MPI
      mcw = mpi_comm_world
      call MPI_INIT (mpi_err)
      if (mpi_err.ne.0) call error_stop(progname, subname, 
     +     'Error in MPI_INIT')
      call MPI_COMM_RANK (mcw, iam, mpi_err)
      if (mpi_err.ne.0) call error_stop(progname, subname, 
     +     'Error in MPI_COMM_RANK')
      call MPI_COMM_SIZE (mcw, num_nodes, mpi_err)
      if (mpi_err.ne.0) call error_stop(progname, subname, 
     +     'Error in MPI_COMM_SIZE')
      if (num_nodes > max_numnodes) then
         write(*, 9001) num_nodes, max_numnodes
 9001    format(//' Requested number of nodes = NUM_NODES    = ',i10/
     +        ' Available number of nodes = MAX_NUMNODES = ',i10/)
         call error_stop(progname, subname,  
     +        'NUM_NODES requested exceeds MAX_NUMNODES')
      endif
#else
      iam = 0
      num_nodes = 1
#endif

      call system_clock(clock1)

      if(iam == 0) then
         write(*,*) 'PCTFR: Parallel CTF refinement'
         write(*,*) 'Parallel version run on ', num_nodes , ' CPUs'
         write(*,*) 'For doc see: http://cryoem.ucsd.edu/programs.shtm'
         write(*,*) 

         oddeven = 0
         call parse_cmd_line(infile)

         if(len_trim(infile) == 0) then
            inunit = 5
            call info(inunit)
         else
            inunit=17
            open(file=infile, unit=inunit)
            call info(inunit)
            close(inunit)
         endif

         open(unit=17, file=in_3d_name, status='old',
     $        form='unformatted', iostat=ios)
         close(17)
         if(ios .ne. 0) then
            write(*,*) 
            write(*,*) 'input map not found: ', trim(in_3d_name)
            call error_stop(progname, subname,
     $           'Unable to open map')
            write(*,*) 
         endif

         nptles = sum(nptle_per_img)
         if(nptles == 0) then
            call error_stop(progname, subname, 
     +           'All particle parameter files are empty')
         endif

         write(*,*)
         write(*,*) 'Number of parameter files = ', nparam
         write(*,*) 'Number of image files     = ', nimg
         write(*,*) 'Total number of particles: ', nptles
         write(*,*)
 
         k = 0
         do i=1,nparam
            do j=1,nimg_per_param(i)
               k = k + 1
               write(*,*) 'param #',i,'orient file: ',
     $              trim(adjustl(param_fnames(i)))
               write(*,*) 'image #',k,'image  file: ',
     $              trim(adjustl(img_fnames(k)))
               write(*,*) 'views = ',  nids(k)
               write(*,*) 'numcols = ',npix_col,'numrows = ',npix_row
               write(*,*)'Ptle records = ', nptle_per_img(k)
               write(*,*)
            enddo
         enddo

         if(ctfmode <=0 .or. ctfmode > 4) then
            write(*,*) 'CTFMODE = ', ctfmode
            call error_stop(progname, subname,
     +           'Program does nothing if ctfmode <= 0 or ctfmode > 4')
         endif

C     Enforce filter to 0, in order not to correct for the CTF (just apply)
         filter = 0
 
C     Check the map, and get the col and row number. 3D model should
C     have the smallest pixel_size and largest box_size compared with
C     boxed images.

         call get_map_dimensions(in_3d_name, npix_col, npix_row, 
     $        npix_sec, angstrom_per_pix, ierr)

         if(ierr > 0) then
            call error_stop(progname, subname,
     +           'Error in 3D map mode, it should be 20 or 40')
         endif

C     3D map should have larger dimension than image box size
         if(npix_row < box_out) then
            call error_stop(progname, subname, 
     $           'NPIX_ROW is less than BOX_OUT')
         endif
         
C     Check the map pixel size
         if(angstrom_per_pix < 0.01 .or. angstrom_per_pix > 100.0) then
            write(*,*) '3D map has nonsensical pixel size'
            write(*,*) 'Use pixel size from particle parameter files'
            angstrom_per_pix = PIXSIZ_OUT
         elseif (abs(angstrom_per_pix - PIXSIZ_OUT) > 10E-5) then
            write(*,*) '3D map pixel size:', angstrom_per_pix
            write(*,*) 'Particle pixel size:', PIXSIZ_OUT
            call error_stop(progname, subname, 
     $       'Pixel size in 3D map and parameter files are different')
         else
            PIXSIZ_OUT = angstrom_per_pix
         endif
      endif

#ifdef MPI
C     Broadcast parameters from root to all other processes
      call bcast_parameters 
#endif
      

C     CALL INTLZ_PARAMS
C     *****************
      call intlz_params(hkl_max, num_hkl, min_xyz, max_xyz, 
     +     nrffta_row, nrffta_col, nrffta_omega, nfc_2d_row, 
     +     nfc_2d_col, nrffts_1_row, nrffts_1_col, 
     +     nrffts_1_omega, nrffts_2_row, nrffts_2_col, 
     +     nrffts_2_omega, den_min_x, den_max_x, den_min_y, 
     +     den_max_y, den_min_z, den_max_z,
     +     confine_3D)

C      if(iam == 0) then
C         print *, ' zero_fill after  call of INTLZ_PARAMS =',zero_fill
C         print *, ' nrffta_row after call of INTLZ_PARAMS =',nrffta_row
C      endif

      call intlz_arrays(hkl_max, min_xyz, max_xyz, min_z, max_z,
     $     first_hk, last_hk, max_hk_slab, first_iq, last_iq,
     $     max_iq_slab, first_rr_out, last_rr_out, max_rr_out_slab,
     $     first_u, last_u, max_u_slab, first_y, last_y, max_y_slab, 
     $     first_z, last_z, max_z_slab, npix_row)


C     Initialize the CTF mem for half fft_size matrix
      
      mem_size = (nfc_2d_row +1) * (nfc_2d_col * 2 +1)
      allocate(ctf_val(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col),
     +     abs_ctf(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col),
     +     disef(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col), 
     +     stat = alloc_stat)
      mem_ctf_val = dble( 12 * mem_size)
      
      if(alloc_stat > 0) then
         call error_stop (progname, subname,
     +      ' Memory allocation failed for ctf_val, abs_ctf, and disef')
      endif

      mem_total = mem_total + mem_ctf_val
      mem_peak  = max(mem_peak, mem_total)

C ---------- End of Initializationl --- Synchronize all the nodes ------

#ifdef MPI
      call mpi_barrier(mcw, mpi_err)
#endif

#ifdef MPI
      call mpi_bcast(param_fnames, filename_len*pfilesmx, mpi_character,
     $     0, mcw, mpi_err)
      call mpi_bcast(img_fnames, filename_len*pfilesmx, mpi_character,
     $     0, mcw, mpi_err)

C     Broadcast number of parameter files, images, particles,
C     images per parameter file, and particles per image
      call mpi_bcast(nparam, 1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(nimg,   1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(nptles, 1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(nimg_per_param, nparam, mpi_integer, 
     $     0, mcw, mpi_err)
      call mpi_bcast(nptle_per_img, nimg, mpi_integer, 
     $     0, mcw, mpi_err)
      call mpi_bcast(first_index, nimg, mpi_integer, 
     $     0, mcw, mpi_err)
#endif

C     Allocate global and local arrays for particle parameters
      allocate(indx(nptles), theta(nptles), phi(nptles), omega(nptles),
     $     x(nptles), y(nptles), mag(nptles), score(nptles),
     $     orient_flag(nptles), origin_flag(nptles))
      allocate(delf_maj_ptle(nptles), delf_min_ptle(nptles),
     $     ang_maj_ptle(nptles))

C     Get particle parameters
C     particle defocuses are not read because each time
C     refinement is done around the global value
      if(iam == 0) then
         call get_ptle_ioom(nparam, filename_len, param_fnames,
     $        oddeven, nptles, nbox, indx, theta, phi, omega, x, y,
     $        mag)
      endif
      
C     Distribute particle information to all other processes
#ifdef MPI
      call mpi_bcast(indx,  nptles, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(theta, nptles, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(phi,   nptles, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(omega, nptles, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(x,     nptles, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(y,     nptles, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(mag,   nptles, mpi_real,    0, mcw, mpi_err)
#endif

      call system_clock(clock2)
      t_init = elapsed_time(clock1, clock2)

C     **********************************************************************
C     *             3D DFT of input 3D density map                         *
C     **********************************************************************

      mem_size = npix_col * npix_row * max_y_slab
      allocate(map3d(npix_col, npix_row, max_y_slab), 
     +     stat = alloc_stat)
      mem_map3d = dble(4 * mem_size) 
      if (alloc_stat > 0) call error_stop (progname, subname,
     +     '3D density map memory allocation failed')
      mem_total = mem_total + mem_map3d
      mem_peak  = max(mem_peak, mem_total)

      call read_map(in_3d_name, map3d, npix_col, npix_row,
     +     max_y_slab, mapprjbig)

C     3d DFT, and exchange 8 quater

      allocate(rffta_omega(nrffta_omega ), stat = alloc_stat)
      if(alloc_stat > 0) call error_stop (progname, subname,
     +     'Error in allocating memory for nrffta_omega')

      call vrffti( nrffta_col, rffta_omega )

      mem_size = (hkl_max+1)* num_hkl * (last_y - first_y + 1) 
      allocate(fft3d(0:hkl_max, -hkl_max:hkl_max, first_y:last_y),
     +  stat = alloc_stat)
      mem_fft3d = dble(8 * mem_size) 
      if(alloc_stat > 0) then
         call error_stop (progname, subname,
     +        '3D density fft map memory allocation failed')
      endif
      mem_total = mem_total + mem_fft3d
      mem_peak  = max(mem_peak, mem_total)

C     FFT3D, (X,Y,Z) real -> (H,K,Z) complex(0:Hmax, Kmin:Kmax, Z_slab
C     Zmin:Zmax) because I change intlz_params.f to make
C     hkl_max = nfc_2d_row = nfc_2d_col

      do i = 0, last_y - first_y
         center(1) = (npix_row-1)/2.0
         center(2) = (npix_col-1)/2.0
         call fft_2dfft(map3d(1,1,i+1), 
     +        npix_row, npix_col, center,
     +        nrffta_omega, rffta_omega, nrffta_row, nrffta_col, 
     +        hkl_max, hkl_max, fft3d(0,-hkl_max,i+first_y))
      enddo

      deallocate(map3d)
      mem_total = mem_total - mem_map3d

C     Exchange data between nodes from Z_slab to H_slab

#ifdef MPI
      if (num_nodes > 1) then
         mem_size = (hkl_max+1)*num_hkl*(last_y - first_y + 1)
         allocate(fft_klh(mem_size), stat = alloc_stat)
         if(alloc_stat > 0) then
            call error_stop (progname, subname,
     +           '3D density fft exchange memory 1 allocation failed')
         endif
         mem_total = mem_total + 8*mem_size 
         mem_peak  = max(mem_peak, mem_total)
         
         call arrange_3d_1(FFT3D, fft_klh, num_nodes, hkl_max,
     +        first_y, last_y, frst_lst_iq)
         
         deallocate(fft3d)
         mem_total = mem_total - mem_fft3d
         mem_fft3d = 8 * mem_size
         
         mem_size = (last_iq-first_iq+1)*num_hkl*npix_row
         allocate(fft3d(first_iq:last_iq, -hkl_max:hkl_max, npix_row),
     +        stat = alloc_stat)
         if(alloc_stat > 0) then
            call error_stop (progname, subname,
     +           '3D density fft exchange memory 2 allocation failed')
         endif
         mem_total = mem_total + 8*mem_size 
         mem_peak  = max(mem_peak, mem_total)
         
         call exch_3d_1(fft_klh, hkl_max, FFT3D, npix_row)
         
         deallocate(fft_klh)
         mem_total = mem_total - mem_fft3d
         mem_fft3d = 8 * mem_size
      endif
#endif

C     1D fft, (H,K,Z)-->(K,L,H) complex(Kmin:Kmax, Lmin:Lmax, H_slab 0:Hmax)

      mem_size = (last_iq-first_iq+1)* num_hkl**2
      allocate(fft_klh(mem_size), stat = alloc_stat)
      if(alloc_stat > 0) then
         call error_stop (progname, subname,
     +        '3D density fft 1dfft allocation failed')
      endif
      mem_total = mem_total +8*mem_size 
      mem_peak  = max(mem_peak, mem_total)

      CALL cfft_1d(FFT3D, fft_klh, hkl_max, npix_row,
     +     nrffta_omega, rffta_omega)

      deallocate(fft3d)
      mem_total = mem_total - mem_fft3d
      mem_fft3d = 8 * mem_size

C     Gather 3Dfft fft_klh in to a full 3D fft map fft3d use up to
C     annulus to save memory.

      fft_ann = int(real(nrffta_row)*PIXSIZ_OUT/res_max) + 1
      fft_ann = min(fft_ann, hkl_max)

      mem_size = (2*fft_ann+1)*(2*fft_ann+1)*(fft_ann+1)
      allocate(fft3d(-fft_ann:fft_ann, -fft_ann:fft_ann, 0:fft_ann),
     +     stat = alloc_stat)
      if(alloc_stat > 0) then
         call error_stop (progname, subname,
     +        '3D fft memory allocation failed')
      endif
      mem_total = mem_total + 8 * mem_size
      mem_peak  = max(mem_peak, mem_total)

C     gather3d has been modified to handle both serial and MPI versions
      call gather3d(fft_klh, hkl_max, fft3d, fft_ann)
      
      deallocate(fft_klh)
      mem_total = mem_total - mem_fft3d
      mem_fft3d = dble(8 * mem_size) 

#ifdef MPI
      call mpi_bcast(mapprjbig, 1, mpi_real, 0, mcw, mpi_err)
#endif
C     ----- End calculation of 3D transform of input map -----

      call system_clock(clock3)
      t_3dfft = elapsed_time(clock2, clock3)

C **********************************************************************
C *             refine CTF of all input files               *
C **********************************************************************

C  loop for all the input parameter files to do cmpt_ort.
      t_analy     = 0
      mem_cmpt_intrps = 0

      k = 1
      do i = 1, nimg

         if(i > 1) then
            k = k + nptle_per_img(i-1)
         endif

         if(iam == 0) then
            write(*,'(A,I6,A,I8,A,I8)') 'image = ',i,
     $        ' - processing particle # ',k,'/',nptles
         endif

C   enforce anastigmatism if requested
         if (anastigmatic==1) then
            delf_majs(i) = 0.5*(delf_majs(i)+delf_mins(i))
            ang_majs(i) =0.
         endif

C   initialize the parameters for one image file.
         npix_row         =  nbox(i)
         npix_col         =  nbox(i)
         max_views        =  nids(i)
         angstrom_per_pix =  pixsizs(i)

         pix_file_name    = img_fnames(i)
         ort_file_name    = param_fnames(i)
         j = first_index(i)
         if(j <= 0) cycle    ! skipping micrographs w/o images
         if (iam == 0) then
            allocate(pixels(npix_col*npix_row))

            call gen_read_pixel_frame(img_fnames(i), pix_file_unit,
     $           j, npix_col, npix_row, 
     $           npix_col, npix_row, 1, pixels, err_code)

            call frame_stats(npix_row, npix_col, pixels,
     $           fmin, fmax, favg, fstd)

            deallocate(pixels)

            enlargefact = mapprjbig/(fstd*sqrt(real(nrffta_col)))
C            print *, 'enlargefact=',enlargefact, fstd, mapprjbig
         endif

#ifdef MPI
         call mpi_bcast(enlargefact, 1, mpi_real,
     +        0, mpi_comm_world, mpi_err)
         if( mpi_err .ne. 0 ) call error_stop(progname,subname,
     +        'Error in calling MPI_BCAST enlargefact')
#endif

C     Initialize arrays containing CTF correction, absolute value of
C     CTF, and radial scaling using Fortran 90 array syntax

         ctf_val = 1.0
         abs_ctf = 1.0
         disef   = 1.0
         
         if (funcweight > 0.0) then 
            call disef_para(disef, nfc_2d_row, nfc_2d_col, 
     $           funcweight, angstrom_per_pix, units)
         endif

         mem_cmpt = 0

         call cmpt_ctf(fft3d, hkl_max, fft_ann,
     $        t_analy, mem_cmpt, ctf_val, abs_ctf,
     $        disef, i, enlargefact, nptle_per_img(i),
     $        indx(k), theta(k), phi(k), omega(k),
     $        x(k), y(k), mag(k), score(k), orient_flag(k),
     $        origin_flag(k), delf_maj_ptle(k),
     $        delf_min_ptle(k), ang_maj_ptle(k))

         mem_cmpt_intrps = max(mem_cmpt_intrps, mem_cmpt)
         mem_peak  = max(mem_peak, mem_total+mem_cmpt)

C     finish read the pif file, Close PIF file.
         if(iam == 0) then
            print *, 'Finished image file: ', pix_file_name
         endif
         
      enddo

C     Write out new particle parameter files from root process
      if(iam == 0) then
         id = 0
         img_id = 0
         do i=1,nparam
            call increment_fname(param_fnames(i), new_name)
            open(unit=11, file=new_name)
            do j=1,nimg_per_param(i)
               img_id = img_id + 1
               write(11,'(A,I6,I6,I8)') trim(img_fnames(img_id)),
     $              nbox(img_id),nbox(img_id),nids(img_id)
               write(11,1010) pixsizs(img_id), units, 
     $              volts(img_id), ampfacs(img_id), 
     $              delf_majs(img_id), delf_mins(img_id),
     $              ang_majs(img_id), cs_coeffs(img_id)
               if(per_ptle_ctf == 1) then
                  do k=1,nptle_per_img(img_id)
                     id = id+1
                     write(11,1235) indx(id), 
     $                    theta(id), phi(id), omega(id),
     $                    x(id), y(id), mag(id), 
     $                    score(id), zero, zero,
     $                    delf_maj_ptle(id), delf_min_ptle(id),
     $                    ang_maj_ptle(id),
     $                    orient_flag(id), origin_flag(id)
                  enddo
               else
                  do k=1,nptle_per_img(img_id)
                     id = id+1
                     write(11,1234) indx(id), 
     $                    theta(id), phi(id), omega(id),
     $                    x(id), y(id), mag(id), 
     $                    score(id), zero, zero,
     $                    orient_flag(id), origin_flag(id)
                  enddo
               endif
            enddo
            close(11)
         enddo
      endif

 1010 format(1X,f8.4,',',i2,',',f10.1,',',4(f10.4,','),f8.2) 
 1234 format(I7,',',4(F8.3,','),F7.3,',',F6.3,',',
     +     F10.3,',',F5.1,',',F5.1,',',1x,A1,',',1x,A1)
 1235 format(I7,',',4(F8.3,','),F7.3,',',F6.3,',',
     +     F10.3,',',F5.1,',',F5.1,',',F7.4,',',F7.4,',',F7.2,',',
     +     1x,A1,',',1x,A1)


      deallocate(ctf_val, abs_ctf, disef, fft3d, indx,
     $     theta, phi, omega, x, y, mag, score, orient_flag,
     $     origin_flag, rffta_omega)

      deallocate(delf_maj_ptle, delf_min_ptle, ang_maj_ptle)

      mem_total = mem_total - mem_ctf_val - mem_fft3d

#ifdef MPI
      call mpi_barrier(mpi_comm_world, mpi_err)
#endif

      call system_clock(clock4)
      t_interp = elapsed_time(clock3, clock4) - t_analy

      if (mem_total .ne. 0) print *,'mem_total=',mem_total,
     +     'At the end, mem_total should be 0!!!'

      if (iam == 0) then
C     Print timing message.
         time = t_init + t_3dfft + t_analy + t_interp
         write(*,40) iam, t_init, t_3dfft, t_analy, t_interp, time
 40      format (/ 'Node ', i3, 
     +        ' Time used in initialization                      =',
     +        f15.4, ' ' / 8x,
     +        ' Time used in read map & 3DFFT                    =',
     +        f15.4, ' ' / 8x,
     +        ' Time used in Read & FFT images & 2D Cut          =',
     +        f15.4, ' ' / 8x,
     +        ' Time used in refine CTF                          =', 
     +        f15.4, ' ' / 8x, 
     +        ' Total time used                                  =',
     +        f15.4)

C     Print memory allocating message.
         write(*,60) iam, mem_peak, mem_map3d, mem_fft3d,
     $        mem_cmpt_intrps
 60      format(/ 'Node ', i3, ' Memory allocation in BYTES:', /
     +        8x, ' Peak memory allocated   =', f18.1 /
     +        8x, ' Memory for 3D Map       =', f18.1 /
     +        8x, ' Memory for 3D FFT       =', f18.1 /
     +        8x, ' Memory used in CMPT_CTF =', f18.1 )

         print *, '**********finished nicely**********'
      endif

#ifdef MPI
      call MPI_FINALIZE ( mpi_err )
#endif

      end program pctfr
