      program pctfr

      implicit none

#ifdef MPI
      include 'mpif.h'
#endif

      include '../include/sizes.inc'
      include '../include/allprogs.inc'
      include '../include/info.inc'
      include '../include/com_01.inc'
      include '../include/mem_time.inc'
      include '../include/pifhead.inc'
      
      logical :: confine_3D = .false.

      character(len=17)  :: subname = 'PCTFR main routine'
      character(len=256) :: infile, new_name

      real :: zero=0.0
      real weight, mapprjbig, enlargefact, center(2)
      real fmin, fmax, favg, fstd
      real :: defocus_max, defocus_min, defocus_ang

      integer min_ctfs(3)
      integer mcw, id, local_max, local_count, buffer_size(0:255)
      integer nptles, img_id, err_code, nview_mod, nomega
      integer fft_ann, i, j, k, inunit, alloc_stat, ierr
      integer first_iq, last_iq, max_iq_slab,
     +     first_z,last_z, first_hk,last_hk, max_hk_slab, num_hkl,
     +     den_min_x, den_max_x, den_min_y, den_max_y, 
     +     den_min_z, den_max_z, mem_size, 
     +     min_xyz, max_xyz, hkl_max, max_z_slab,
     +     first_rr_out, last_rr_out, max_rr_out_slab,
     +     first_u, last_u, max_u_slab,
     +     first_y, last_y, max_y_slab
      integer nrffts_1_row, nrffts_1_col, nrffts_1_omega,
     +     nrffts_2_row, nrffts_2_col, nrffts_2_omega
      
      double precision elapsed_time

      integer clock1, clock2, ios
      integer nsize

      integer, allocatable   :: indx(:)
      real, allocatable      :: pixels(:), theta(:), phi(:), omega(:)
      real, allocatable      :: x(:), y(:), mag(:), score(:)
      real, allocatable      :: delf_maj_ptle(:), delf_min_ptle(:)
      real, allocatable      :: ang_maj_ptle(:)
      real, allocatable      :: ctf_val(:,:), abs_ctf(:,:), disef(:,:)
      real, allocatable      :: map3d(:,:,:), rffta_omega(:)
      real, allocatable      :: real_results_local(:,:)
      character, allocatable :: orient_flag(:), origin_flag(:)
      character, allocatable :: char_results_local(:,:)
      complex, allocatable   :: fft3d(:,:,:), fft_klh(:)

      real, allocatable      :: ctfr_score(:,:,:)        ! array of ctf scores per particle
      real, allocatable      :: sum_score(:,:,:)         ! array of summed ctf scores
      integer, allocatable   :: n_score(:,:,:)           ! count of valid summed ctf scores
      real, allocatable      :: local_sum_score(:,:,:)   ! array of locally summed ctf scores
      integer, allocatable   :: local_n_score(:,:,:)     ! count of locallyvalid summed ctf scores
      real, allocatable      :: ctfr_results_local(:,:)  ! local array of ctf values per particle

#ifdef MPI
      integer stat(mpi_status_size)
#endif

      PROGNAME = 'PCTFR '

#ifdef MPI
      mcw = mpi_comm_world
      call MPI_INIT (mpi_err)
      if (mpi_err.ne.0) call error_stop(progname, subname, 
     +     'Error in MPI_INIT')
      call MPI_COMM_RANK (mcw, iam, mpi_err)
      if (mpi_err.ne.0) call error_stop(progname, subname, 
     +     'Error in MPI_COMM_RANK')
      call MPI_COMM_SIZE (mcw, num_nodes, mpi_err)
      if (mpi_err.ne.0) call error_stop(progname, subname, 
     +     'Error in MPI_COMM_SIZE')
      if (num_nodes > max_numnodes) then
         write(*, 9001) num_nodes, max_numnodes
 9001    format(//' Requested number of nodes = NUM_NODES    = ',i10/
     +        ' Available number of nodes = MAX_NUMNODES = ',i10/)
         call error_stop(progname, subname,  
     +        'NUM_NODES requested exceeds MAX_NUMNODES')
      endif
      if (num_nodes > 255) then
         write(*, 9002) num_nodes
 9002    format(//' Requested number of nodes = NUM_NODES    = ',i10/
     +        ' Maximum allowed because of internal buufer size = 255'/)
         call error_stop(progname, subname,  
     +        'NUM_NODES requested exceeds internal buffer size')
      endif
#else
      iam = 0
      num_nodes = 1
#endif

      call system_clock(clock1)

      if(iam == 0) then
         write(*,*) 'PCTFR: Parallel CTF refinement'
         write(*,*) 'Parallel version run on ', num_nodes , ' CPUs'
         write(*,*) 'For doc see: http://cryoem.ucsd.edu/programs.shtm'
         write(*,*) 

         oddeven = 0
         call parse_cmd_line(infile)

         if(len_trim(infile) == 0) then
            inunit = 5
            call info(inunit)
         else
            inunit=17
            open(file=infile, unit=inunit)
            call info(inunit)
            close(inunit)
         endif

         open(unit=17, file=in_3d_name, status='old',
     $        form='unformatted', iostat=ios)
         close(17)
         if(ios .ne. 0) then
            write(*,*) 
            write(*,*) 'input map not found: ', trim(in_3d_name)
            call error_stop('PCTFR', 'Main program',
     $           'Unable to open map')
            write(*,*) 
         endif

         nptles = sum(nptle_per_img)
         if(nptles == 0) then
            call error_stop('PCTFR', 'main program', 
     +           'All particle parameter files are empty')
         endif

         write(*,*)
         write(*,*) 'Number of parameter files = ', nparam
         write(*,*) 'Number of image files     = ', nimg
         write(*,*) 'Total number of particles: ', nptles
         write(*,*)
         
         write(*,*)
         write(*,*) 'Defocus refinement: (To add further information)'
         write(*,*)

         k = 0
         do i=1,nparam
            do j=1,nimg_per_param(i)
               k = k + 1
               write(*,*) 'param #',i,'orient file: ',
     $              trim(adjustl(param_fnames(i)))
               write(*,*) 'image #',k,'image  file: ',
     $              trim(adjustl(img_fnames(k)))
               write(*,*) 'views = ',  nids(k)
               write(*,*) 'numcols = ',npix_col,'numrows = ',npix_row
               write(*,*)'Ptle records = ', nptle_per_img(k)
               write(*,*)
            enddo
         enddo

         if(ctfmode <=0 .or. ctfmode > 4) then
            write(*,*) 'CTFMODE = ', ctfmode
            call error_stop(progname, subname,
     +           'Program does nothing if ctfmode <= 0 or ctfmode > 4')
         endif
         
C     Check the map, and get the col and row number. 3D model should
C     have the smallest pixel_size and largest box_size compared with
C     boxed images.

         call get_map_dimensions(in_3d_name, npix_col, npix_row, 
     $        npix_sec, angstrom_per_pix, ierr)
         if(ierr > 0) then
            call error_stop('PCTFR', 'PCTFR main',
     +           'Error in 3D map mode, it should be 20 or 40')
         endif

C     3D map should have larger dimension than image box size
         if(npix_row < box_out) then
            call error_stop(progname, subname, 
     $           'NPIX_ROW is less than BOX_OUT')
         endif
         
C     Check the map pixel size
         if(angstrom_per_pix < 0.01 .or. angstrom_per_pix > 100.0) then
            write(*,*) '3D map has nonsensical pixel size'
            write(*,*) 'Use pixel size from particle parameter files'
            angstrom_per_pix = PIXSIZ_OUT
         elseif (abs(angstrom_per_pix - PIXSIZ_OUT) > 10E-5) then
            write(*,*) '3D map pixel size:', angstrom_per_pix
            write(*,*) 'Particle pixel size:', PIXSIZ_OUT
            call error_stop(progname, subname, 
     $       'Pixel size in 3D map and parameter files are different')
         else
            PIXSIZ_OUT = angstrom_per_pix
         endif
         
      endif

#ifdef MPI
C     Broadcast parameters from root to all other processes
      call bcast_parameters 
#endif

      call intlz_params(hkl_max, num_hkl, min_xyz, max_xyz, 
     +     nrffta_row, nrffta_col, nrffta_omega, nfc_2d_row, 
     +     nfc_2d_col, nrffts_1_row, nrffts_1_col, 
     +     nrffts_1_omega, nrffts_2_row, nrffts_2_col, 
     +     nrffts_2_omega, den_min_x, den_max_x, den_min_y, 
     +     den_max_y, den_min_z, den_max_z,
     +     confine_3D)

      if(iam == 0) then
         print *, ' zero_fill after  call of INTLZ_PARAMS =',zero_fill
         print *, ' nrffta_row after call of INTLZ_PARAMS =',nrffta_row
      endif

      call intlz_arrays(hkl_max, min_xyz, max_xyz, min_z, max_z,
     +     first_hk, last_hk, max_hk_slab, first_iq, last_iq,
     +     max_iq_slab, first_rr_out, last_rr_out, max_rr_out_slab,
     +     first_u, last_u, max_u_slab,
     +     first_y, last_y, max_y_slab, first_z, last_z, max_z_slab,
     +     npix_row)

C     Initialize the CTF mem for half fft_size matrix
      
      mem_size = (nfc_2d_row +1) * (nfc_2d_col * 2 +1)
      allocate(ctf_val(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col),
     +     abs_ctf(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col),
     +     disef(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col), 
     +     stat = alloc_stat)
      mem_ctf_val = dble( 12 * mem_size)
      
      if(alloc_stat > 0) then
         call error_stop (progname, subname,
     +      ' Memory allocation failed for ctf_val, abs_ctf, and disef')
      endif

      mem_total = mem_total + mem_ctf_val
      mem_peak  = max(mem_peak, mem_total)

      mem_cmpt_intrps = 0

C     Broadcast parameter file names and image file names

#ifdef MPI
      call mpi_bcast(param_fnames, filename_len*pfilesmx, mpi_character,
     $     0, mcw, mpi_err)
      call mpi_bcast(img_fnames, filename_len*pfilesmx, mpi_character,
     $     0, mcw, mpi_err)
C
C     Broadcast number of parameter files, images, particles,
C     images per parameter file, and particles per image
C
      call mpi_bcast(nparam, 1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(nimg,   1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(nptles, 1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(nimg_per_param, nparam, mpi_integer, 
     $     0, mcw, mpi_err)
      call mpi_bcast(nptle_per_img, nimg, mpi_integer, 
     $     0, mcw, mpi_err)
      call mpi_bcast(first_index, nimg, mpi_integer, 
     $     0, mcw, mpi_err)
#endif

C     Calculate buffer sizes
      buffer_size = 0
      do i=0,num_nodes-1
         buffer_size(i) = nptles/num_nodes
         if(i < mod(nptles,num_nodes)) then
            buffer_size(i) = buffer_size(i) + 1
         endif
      enddo

C     Calculate the maximum number of particles assigned to a process
      local_max = maxval(buffer_size)

C     Allocate global and local arrays for particle parameters
      allocate(indx(nptles), theta(nptles), phi(nptles), omega(nptles),
     $     x(nptles), y(nptles), mag(nptles), score(nptles),
     $     orient_flag(nptles), origin_flag(nptles))
#ifdef MPI
      allocate(real_results_local(1,local_max),
     $     char_results_local(2,local_max))

#endif      
 
C     Allocate arrays for defocus refinement
      allocate(ctfr_score
     $       (-Nangle:Nangle,-Nangle:Nangle,-Nangle:Nangle))
      if (per_ptle_ctf == 0) then
         allocate(sum_score
     $       (-Nangle:Nangle,-Nangle:Nangle,-Nangle:Nangle))
         allocate(n_score
     $       (-Nangle:Nangle,-Nangle:Nangle,-Nangle:Nangle))
         allocate(local_sum_score
     $       (-Nangle:Nangle,-Nangle:Nangle,-Nangle:Nangle))
         allocate(local_n_score
     $       (-Nangle:Nangle,-Nangle:Nangle,-Nangle:Nangle))
      endif 
#ifdef MPI
      if(per_ptle_ctf == 1) then
         allocate(ctfr_results_local(3, local_max))
      endif
#endif      

      if(per_ptle_ctf == 1) then
         allocate(delf_maj_ptle(nptles), delf_min_ptle(nptles),
     $        ang_maj_ptle(nptles))
      endif

C     Get particle parameters
      if(iam == 0) then
         if(per_ptle_ctf == 1) then
            call get_ptle_ioomctf(nparam, filename_len, param_fnames,
     $           oddeven, nptles, nbox, indx, theta, phi, omega, x, y,
     $           mag, delf_maj_ptle, delf_min_ptle, ang_maj_ptle)
         else                  
            call get_ptle_ioom(nparam, filename_len, param_fnames,
     $           oddeven, nptles, nbox, indx, theta, phi, omega, x, y,
     $           mag)
         endif
      endif

C     Distribute particle information to all other processes
#ifdef MPI
      call mpi_bcast(indx,  nptles, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(theta, nptles, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(phi,   nptles, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(omega, nptles, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(x,     nptles, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(y,     nptles, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(mag,   nptles, mpi_real,    0, mcw, mpi_err)
      if(per_ptle_ctf == 1) then
         call mpi_bcast(delf_maj_ptle, nptles, mpi_real, 0,mcw,mpi_err)
         call mpi_bcast(delf_min_ptle, nptles, mpi_real, 0,mcw,mpi_err)
         call mpi_bcast(ang_maj_ptle,  nptles, mpi_real, 0,mcw,mpi_err)
      endif
#endif

      if(iam==0) write(*,*) 'I am starting!', per_ptle_ctf
C     ----- Start calculation of 3D transform of input map -----
      mem_size = npix_col * npix_row * max_y_slab
      allocate(map3d(npix_col, npix_row, max_y_slab), 
     +     stat = alloc_stat)
      mem_map3d = dble(4 * mem_size) 
      if (alloc_stat > 0) call error_stop (progname, subname,
     +     '3D density map memory allocation failed')
      mem_total = mem_total + mem_map3d
      mem_peak  = max(mem_peak, mem_total)
      
      call read_map(in_3d_name, map3d, npix_col, npix_row,
     +     max_y_slab, mapprjbig)

C     3d DFT, and exchange 8 quater

      allocate(rffta_omega(nrffta_omega ), stat = alloc_stat)
      if(alloc_stat > 0) call error_stop (progname, subname,
     +     'Error in allocating memory for nrffta_omega')
      
      call vrffti( nrffta_col, rffta_omega )
      
      mem_size = (hkl_max+1)* num_hkl * (last_y - first_y + 1) 
      allocate(fft3d(0:hkl_max, -hkl_max:hkl_max, first_y:last_y),
     +     stat = alloc_stat)
      mem_fft3d = dble(8 * mem_size) 
      if(alloc_stat > 0) then
         call error_stop (progname, subname,
     +        '3D density fft map memory allocation failed')
      endif
      mem_total = mem_total + mem_fft3d
      mem_peak  = max(mem_peak, mem_total)

C     FFT3D, (X,Y,Z) real -> (H,K,Z) complex(0:Hmax, Kmin:Kmax, Z_slab
C     Zmin:Zmax) because I change intlz_params.f to make hkl_max =
C     nfc_2d_row = nfc_2d_col

      do i = 0, last_y - first_y
         center(1) = (npix_row-1)/2.0
         center(2) = (npix_col-1)/2.0
         call fft_2dfft(map3d(1,1,i+1), 
     +        npix_row, npix_col, center,
     +        nrffta_omega, rffta_omega, nrffta_row, nrffta_col, 
     +        hkl_max, hkl_max, fft3d(0,-hkl_max,i+first_y))
      enddo

      deallocate(map3d)
      mem_total = mem_total - mem_map3d

#ifdef MPI
      call mpi_barrier(mcw, mpi_err)
#endif

C     Exchange data between nodes from Z_slab to H_slab

#ifdef MPI
      if (num_nodes > 1) then   
         mem_size = (hkl_max+1)*num_hkl*(last_y - first_y + 1)
         allocate(fft_klh(mem_size), stat = alloc_stat)
         if(alloc_stat > 0) then
            call error_stop (progname, subname,
     +           '3D density fft exchange memory 1 allocation failed')
         endif
         mem_total = mem_total + 8*mem_size 
         mem_peak  = max(mem_peak, mem_total)
         
         call arrange_3d_1(FFT3D, fft_klh, num_nodes, hkl_max,
     +        first_y, last_y, frst_lst_iq)
         
         deallocate(fft3d)
         mem_total = mem_total - mem_fft3d
         mem_fft3d = 8 * mem_size
         
         mem_size = (last_iq-first_iq+1)*num_hkl*npix_row
         allocate(fft3d(first_iq:last_iq, -hkl_max:hkl_max, npix_row),
     +        stat = alloc_stat)
         if(alloc_stat > 0) then
            call error_stop (progname, subname,
     +           '3D density fft exchange memory 2 allocation failed')
         endif
         mem_total = mem_total + 8*mem_size 
         mem_peak  = max(mem_peak, mem_total)
         
         call exch_3d_1(fft_klh, hkl_max, FFT3D, npix_row)

         deallocate(fft_klh)
         mem_total = mem_total - mem_fft3d
         mem_fft3d = 8 * mem_size
      endif
#endif

Cji   1D fft, (H,K,Z)-->(K,L,H) complex(Kmin:Kmax, Lmin:Lmax, H_slab 0:Hmax)

      mem_size = (last_iq-first_iq+1)* num_hkl**2
      allocate(fft_klh(mem_size), stat = alloc_stat)
      if(alloc_stat > 0) then
         call error_stop (progname, subname,
     +        '3D density fft 1dfft allocation failed')
      endif
      mem_total = mem_total +8*mem_size 
      mem_peak  = max(mem_peak, mem_total)

      CALL cfft_1d(FFT3D, fft_klh, hkl_max, npix_row,
     +     nrffta_omega, rffta_omega)

      deallocate(fft3d)
      mem_total = mem_total - mem_fft3d
      mem_fft3d = 8 * mem_size

C     Gather 3Dfft fft_klh in to a full 3D fft map fft3d use up to
C     annulus to save memory.
      fft_ann = int(real(nrffta_row)*PIXSIZ_OUT/res_max) + 1
      fft_ann = min(fft_ann, hkl_max)

      mem_size = (2*fft_ann+1)*(2*fft_ann+1)*(fft_ann+1)
      allocate(fft3d(-fft_ann:fft_ann, -fft_ann:fft_ann, 0:fft_ann),
     +     stat = alloc_stat)
      if(alloc_stat > 0) then
         call error_stop (progname, subname,
     +        '3D fft memory allocation failed')
      endif
      mem_total = mem_total + 8 * mem_size
      mem_peak  = max(mem_peak, mem_total)

C     gather3d has been modified to handle both serial and MPI versions
      call gather3d(fft_klh, hkl_max, fft3d, fft_ann)
      
      deallocate(fft_klh)
      mem_total = mem_total - mem_fft3d
      mem_fft3d = dble(8 * mem_size) 

#ifdef MPI
      call mpi_bcast(mapprjbig, 1, mpi_real, 0, mcw, mpi_err)
#endif
C     ----- End calculation of 3D transform of input map -----

      if(iam==0) write(*,*) 'I am DONE!'
      id = 0
      local_count = 0
      do i= 1, nimg
            
C         write(*,*) 'image',i,'of',nimg
         npix_row         =  nbox(i)
         npix_col         =  nbox(i)
         angstrom_per_pix = pixsizs(i)
         
            
         j = first_index(i)
         if(j <= 0) cycle    ! skipping micrographs w/o images

         allocate(pixels(npix_col*npix_row))
         if (iam == 0) then

            call gen_read_pixel_frame(img_fnames(i), pix_file_unit,
     $           j, npix_col, npix_row, 
     $           npix_col, npix_row, 1, pixels, err_code)

            call frame_stats(npix_row, npix_col, pixels,
     $           fmin, fmax, favg, fstd)


            enlargefact = mapprjbig/(fstd*sqrt(real(nrffta_col)))
         endif

#ifdef MPI
         call mpi_bcast(enlargefact, 1, mpi_real,
     +        0, mcw, mpi_err)
C         if( mpi_err .ne. 0 ) call error_stop(progname,subname,
C     +        'Error in calling MPI_BCAST enlargefact')
#endif

C     Initialize arrays containing CTF correction, absolute value of
C     CTF, and radial scaling using Fortran 90 array syntax

         ctf_val = 1.0
         abs_ctf = 1.0
         disef   = 1.0
         
         if (funcweight > 0.0) then 
            call disef_para(disef, nfc_2d_row, nfc_2d_col, 
     $           funcweight, angstrom_per_pix, units)
         endif
            
         if (ctfmode > 0 .and. ctfmode <= 4)  then
            filter = 0       ! Required so that model multiplied by CTF
            weight = 1.0/enlargefact
         endif

         do j=1, nptle_per_img(i)
            id = id + 1

            if (per_ptle_ctf == 1 ) then
               defocus_max = delf_maj_ptle(id)
               defocus_min = delf_min_ptle(id)
               defocus_ang = ang_maj_ptle(id)
            else
               defocus_max = delf_majs(i)
               defocus_min = delf_mins(i)
               defocus_ang = ang_majs(i)
               local_sum_score = 0.
               local_n_score = 0
            endif

            if(iam == mod(id-1,num_nodes)) then
               local_count = local_count + 1
                  
C               write(*,*) 'particle',j,'of',nptle_per_img(i)
               call gen_read_pixel_frame(img_fnames(i), 
     $              pix_file_unit, indx(id), npix_col, npix_row, 
     $              npix_col, npix_row, 1, pixels, err_code)
                  
               if(err_code > 0) then
                  write(*,*) ' --------- WARNING ---------'
                  write(*,*) 'Invalid image'
                  write(*,*) 'Skipping image number ', indx(id)
                  write(*,*) 'Micrograph ', trim(img_fnames(i))
                  write(*,*) 
                  ctfr_score(:,:,:) = 999.99       !F90 array syntax 
                  score(id) = 999.99
                  orient_flag(id) = 'X'
                  origin_flag(id) = 'X'
               else

                  call cmpt_ctf(fft3d, hkl_max, fft_ann, ctf_val,
     $                 abs_ctf, disef, i, enlargefact, nptle_per_img(i),
     $                 pixels, mag(id), theta(id), phi(id), omega(id),
     $                 x(id), y(id), 
     $                 defocus_max, defocus_min, defocus_ang,
     $                 ctfr_score,
     $                 orient_flag(id), origin_flag(id))

                  score(id) = minval(ctfr_score)
                  if (per_ptle_ctf == 1 ) then
                      delf_maj_ptle(id) = defocus_max
                      delf_min_ptle(id) = defocus_min
                      ang_maj_ptle(id)  = defocus_ang
                  else
                      where (ctfr_score < 999.)
                         local_sum_score = local_sum_score + ctfr_score
                         local_n_score = local_n_score + 1
                      endwhere
                  endif            
               endif
                  
#ifdef MPI
               real_results_local(1,local_count) = score(id)
               char_results_local(1,local_count) = orient_flag(id)
               char_results_local(2,local_count) = origin_flag(id)

               
               if (per_ptle_ctf == 1 ) then
                  ctfr_results_local(1,local_count) = defocus_max
                  ctfr_results_local(2,local_count) = defocus_min
                  ctfr_results_local(3,local_count) = defocus_ang
               endif
#endif
            endif
         enddo
            
         deallocate(pixels)

C determine optimum global defocus value
C only if not searching on particle basis
         if (per_ptle_ctf == 0 ) then
#ifdef MPI
           nsize    = (2*Nangle+1)*(2*Nangle+1)*(2*Nangle+1) 
           call MPI_Reduce(local_sum_score, sum_score, nsize,
     +        MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD, mpi_err)
           call MPI_Reduce(local_n_score, n_score, nsize,
     +        MPI_INTEGER, MPI_SUM, 0, MPI_COMM_WORLD, mpi_err)
#else
           sum_score = local_sum_score       ! Fortran 90 array syntax
           n_score = local_n_score           ! Fortran 90 array syntax
#endif

           if (iam==0) then
             sum_score = sum_score / n_score
      
C     Determine location of minimum element in array sum_score. Note use
C     of Fortran 90 array syntax and intrinsic function minloc (which
C     behaves as if lower bounds of array are all equal to 1)

             min_ctfs = minloc(sum_score)
             min_ctfs = min_ctfs - nangle - 1

             delf_majs(i) = delf_majs(i) + min_ctfs(1) * Dangle
             delf_mins(i) = delf_mins(i) + min_ctfs(2) * Dangle
             ang_majs(i)  = ang_majs(i)  + min_ctfs(3) * Dangle * 10
           endif
         endif

      enddo

C     Collect results back onto root process
#ifdef MPI
      do i=1,num_nodes-1
         if(iam == i) then
            call mpi_send(real_results_local, 1*buffer_size(i),
     $           mpi_real, 0, i, mcw, mpi_err)
            call mpi_send(char_results_local, 2*buffer_size(i),
     $           mpi_character, 0, i, mcw, mpi_err)
            if (per_ptle_ctf == 1) then
               call mpi_send(ctfr_results_local,
     $              3*buffer_size(i),
     $              mpi_real, 0, i, mcw, mpi_err)
            endif
         endif
         if(iam == 0) then
            call mpi_recv(real_results_local, 1*buffer_size(i),
     $           mpi_real, i, i, mcw, stat, mpi_err)
            call mpi_recv(char_results_local, 2* buffer_size(i),
     $           mpi_character, i, i, mcw, stat, mpi_err)
            if (per_ptle_ctf == 1) then
               call mpi_recv(ctfr_results_local,
     $              3*buffer_size(i),
     $              mpi_real, i, i, mcw, stat, mpi_err)
            endif
               
            do j = 1, buffer_size(i)
               k = i+1+(j-1)*num_nodes
               score(k)       = real_results_local(1,j)
               orient_flag(k) = char_results_local(1,j)
               origin_flag(k) = char_results_local(2,j)

               if (per_ptle_ctf == 1) then
                  delf_maj_ptle(k) = ctfr_results_local(1,j)
                  delf_min_ptle(k) = ctfr_results_local(2,j)
                  ang_maj_ptle(k)  = ctfr_results_local(3,j)
               endif
            enddo
         endif
      enddo 
#endif

C     Write out new particle parameter files from root process
      if(iam == 0) then
         id = 0
         img_id = 0
         do i=1,nparam
            call increment_fname(param_fnames(i), new_name)
            open(unit=11, file=new_name)
            do j=1,nimg_per_param(i)
               img_id = img_id + 1
               write(11,'(A)') trim(img_fnames(img_id))
               write(11,1010) pixsizs(img_id), units, 
     $              volts(img_id), ampfacs(img_id), 
     $              delf_majs(img_id), delf_mins(img_id),
     $              ang_majs(img_id), cs_coeffs(img_id)
               if(per_ptle_ctf == 1) then
                  do k=1,nptle_per_img(img_id)
                     id = id+1
                     write(11,1235) indx(id), 
     $                    theta(id), phi(id), omega(id),
     $                    x(id), y(id), mag(id), 
     $                    score(id), zero, zero,
     $                    delf_maj_ptle(id), delf_min_ptle(id),
     $                    ang_maj_ptle(id),
     $                    orient_flag(id), origin_flag(id)
                  enddo
               else
                  do k=1,nptle_per_img(img_id)
                     id = id+1
                     write(11,1234) indx(id), 
     $                    theta(id), phi(id), omega(id),
     $                    x(id), y(id), mag(id), 
     $                    score(id), zero, zero,
     $                    orient_flag(id), origin_flag(id)
                  enddo
               endif
            enddo
            close(11)
         enddo
      endif

 1010 format(1X,f8.4,',',i2,',',f10.1,',',4(f10.4,','),f8.2) 
 1234 format(I7,',',4(F8.3,','),F7.3,',',F6.3,',',
     +     F10.3,',',F5.1,',',F5.1,',',1x,A1,',',1x,A1)
 1235 format(I7,',',4(F8.3,','),F7.3,',',F6.3,',',
     +     F10.3,',',F5.1,',',F5.1,',',F7.4,',',F7.4,',',F7.2,',',
     +     1x,A1,',',1x,A1)

      deallocate(ctf_val, abs_ctf, disef, fft3d, indx,
     $     theta, phi, omega, x, y, mag, score, orient_flag,
     $     origin_flag, rffta_omega)
#ifdef MPI
      deallocate(real_results_local, char_results_local)
#endif

      deallocate(ctfr_score)
      if(per_ptle_ctf == 1) then
         deallocate(delf_maj_ptle, delf_min_ptle, ang_maj_ptle)
#ifdef MPI
         deallocate(ctfr_results_local)
#endif      
      else
         deallocate(sum_score, n_score, local_sum_score, local_n_score)
      endif

      call system_clock(clock2)
      time = elapsed_time(clock1, clock2)

      if(iam == 0) then
         write(*,*) 'Total run time = ', time
      endif

#ifdef MPI
      call mpi_finalize(mpi_err)
#endif

      end program pctfr
      